@@ -1,4 +1,6 @@
 {-# OPTIONS_GHC -fno-warn-missing-methods #-}
+{-# LANGUAGE StrictData            #-}
+{-# LANGUAGE BangPatterns          #-}
 {-# LANGUAGE CPP                   #-}
 {-# LANGUAGE DeriveFunctor         #-}
 {-# LANGUAGE FlexibleInstances     #-}
@@ -12,6 +14,7 @@ module Skylighting.Tokenizer (
   ) where
 
 import Control.Applicative
+import Control.Monad
 import Control.Monad.Except
 import Control.Monad.Reader
 import Control.Monad.State.Strict
@@ -37,7 +40,11 @@ import Data.List.NonEmpty (NonEmpty((:|)), (<|), toList)
 import Data.Semigroup
 #endif
 
-newtype ContextStack = ContextStack{ unContextStack :: NonEmpty Context }
+newtype Captures = Captures{ unCaptures :: IntMap.IntMap ByteString }
+  deriving (Show)
+
+newtype ContextStack =
+  ContextStack{ unContextStack :: NonEmpty (Context, Captures) }
   deriving (Show)
 
 data TokenizerState = TokenizerState{
@@ -45,11 +52,10 @@ data TokenizerState = TokenizerState{
   , endline             :: Bool
   , prevChar            :: Char
   , contextStack        :: ContextStack
-  , captures            :: IntMap.IntMap ByteString
+  , captures            :: Captures
   , column              :: Int
   , lineContinuation    :: Bool
   , firstNonspaceColumn :: Maybe Int
-  , compiledRegexes     :: Map.Map RE Regex
 }
 
 -- | Configuration options for 'tokenize'.
@@ -65,9 +71,9 @@ data Result e a = Success a
 
 deriving instance (Show a, Show e) => Show (Result e a)
 
-data TokenizerM a = TM { runTokenizerM :: TokenizerConfig
-                                       -> TokenizerState
-                                       -> (TokenizerState, Result String a) }
+newtype TokenizerM a = TM { runTokenizerM :: TokenizerConfig
+                                          -> TokenizerState
+                                          -> (TokenizerState, Result String a) }
 
 mapsnd :: (a -> b) -> (c, a) -> (c, b)
 mapsnd f (x, y) = (x, f y)
@@ -118,7 +124,7 @@ instance MonadPlus TokenizerM where
 
 instance MonadReader TokenizerConfig TokenizerM where
   ask = TM (\c s -> (s, Success c))
-  local f (TM x) = TM (\c s -> x (f c) s)
+  local f (TM x) = TM (x . f)
 
 instance MonadState TokenizerState TokenizerM where
   get = TM (\_ s -> (s, Success s))
@@ -133,7 +139,7 @@ instance MonadError String TokenizerM where
 -- | Tokenize some text using 'Syntax'.
 tokenize :: TokenizerConfig -> Syntax -> Text -> Either String [SourceLine]
 tokenize config syntax inp =
-  eitherStack >>= \stack ->
+  eitherStack >>= \(!stack) ->
     case runTokenizerM action
             config{ syntaxMap = Map.map (resolveKeywords (syntaxMap config))
                                           (syntaxMap config) }
@@ -145,18 +151,17 @@ tokenize config syntax inp =
     action = mapM tokenizeLine (zip (BS.lines (encodeUtf8 inp)) [1..])
     eitherStack = case lookupContext (sStartingContext syntax)
                          (resolveKeywords (syntaxMap config) syntax) of
-                    Just c  -> Right $ ContextStack (c :| [])
+                    Just c  -> Right $ ContextStack ((c, Captures mempty) :| [])
                     Nothing -> Left "No starting context specified"
     startingState stack =
       TokenizerState{ input = BS.empty
                     , endline = Text.null inp
                     , prevChar = '\n'
                     , contextStack = stack
-                    , captures = mempty
+                    , captures = Captures mempty
                     , column = 0
                     , lineContinuation = False
                     , firstNonspaceColumn = Nothing
-                    , compiledRegexes = Map.empty
                     }
 
 info :: String -> TokenizerM ()
@@ -169,7 +174,7 @@ infoContextStack = do
   tr <- asks traceOutput
   when tr $ do
     ContextStack stack <- gets contextStack
-    info $ "CONTEXT STACK " ++ show (map cName $ toList stack)
+    info $ "CONTEXT STACK " ++ show (map (cName . fst) $ toList stack)
 
 popContextStack :: TokenizerM ()
 popContextStack = do
@@ -184,25 +189,48 @@ pushContextStack :: Context -> TokenizerM ()
 pushContextStack cont = do
   modify (\st -> st{ contextStack =
                       ContextStack
-                       ((cont <|) . unContextStack $ contextStack st) } )
+                       (((cont, Captures mempty) <|) . unContextStack
+                         $ contextStack st) } )
   infoContextStack
 
 currentContext :: TokenizerM Context
 currentContext = do
-  ContextStack (c :| _) <- gets contextStack
+  ContextStack ((c,_) :| _) <- gets contextStack
   return c
 
 doContextSwitch :: ContextSwitch -> TokenizerM ()
 doContextSwitch Pop = popContextStack
-doContextSwitch (Push (syn,c)) = do
+doContextSwitch (Push (!syn,!c)) = do
   syntaxes <- asks syntaxMap
   case Map.lookup syn syntaxes >>= lookupContext c of
-       Just con -> pushContextStack con
-       Nothing  -> throwError $ "Unknown syntax or context: " ++ show (syn, c)
+       Just !con -> pushContextStack con
+       Nothing   -> throwError $ "Unknown syntax or context: " ++ show (syn, c)
 
 doContextSwitches :: [ContextSwitch] -> TokenizerM ()
 doContextSwitches = mapM_ doContextSwitch
 
+addCaptures :: TokenizerM ()
+addCaptures = do
+  capts <- gets captures
+  if IntMap.null (unCaptures capts)
+     then return ()
+     else do
+       ContextStack ((c,_) :| cs) <- gets contextStack
+       info $ "Adding captures to " <> show (cName c) <> ": " <> show capts
+       modify $ \st -> st{ contextStack = ContextStack ((c,capts) :| cs) }
+
+getCapture :: Int -> TokenizerM Text
+getCapture capnum = do
+  ContextStack ((_,Captures capts) :| _) <- gets contextStack
+  info $ "Retrieving capture " <> show capnum
+  res <- case IntMap.lookup capnum capts of
+          Nothing -> do
+            info "Not found"
+            mzero
+          Just x  -> decodeBS x
+  info $ "Got " <> show res
+  return res
+
 lookupContext :: Text -> Syntax -> Maybe Context
 lookupContext name syntax | Text.null name =
   if Text.null (sStartingContext syntax)
@@ -211,16 +239,16 @@ lookupContext name syntax | Text.null name =
 lookupContext name syntax = Map.lookup name $ sContexts syntax
 
 tokenizeLine :: (ByteString, Int) -> TokenizerM [Token]
-tokenizeLine (ln, linenum) = do
+tokenizeLine (!ln, !linenum) = do
   modify $ \st -> st{ input = ln, endline = BS.null ln, prevChar = '\n' }
   cur <- currentContext
   lineCont <- gets lineContinuation
   if lineCont
      then modify $ \st -> st{ lineContinuation = False }
      else do
+       let !mbFirstNonspace = BS.findIndex (not . isSpace) $! ln
        modify $ \st -> st{ column = 0
-                         , firstNonspaceColumn =
-                              BS.findIndex (not . isSpace) ln }
+                         , firstNonspaceColumn = mbFirstNonspace }
        doContextSwitches (cLineBeginContext cur)
   if BS.null ln
      then doContextSwitches (cLineEmptyContext cur)
@@ -240,7 +268,7 @@ getToken :: TokenizerM (Maybe Token)
 getToken = do
   inp <- gets input
   gets endline >>= guard . not
-  context <- currentContext
+  !context <- currentContext
   msum (map (\r -> tryRule r inp) (cRules context)) <|>
      case cFallthroughContext context of
            [] | cFallthrough context -> Nothing <$ doContextSwitches [Pop]
@@ -257,7 +285,7 @@ takeChars numchars = do
   inp <- gets input
   let (bs,rest) = UTF8.splitAt numchars inp
   guard $ not (BS.null bs)
-  t <- decodeBS bs
+  !t <- decodeBS bs
   modify $ \st -> st{ input = rest,
                       endline = BS.null rest,
                       prevChar = Text.last t,
@@ -273,14 +301,17 @@ tryRule rule inp = do
        Just n  -> gets column >>= guard . (== n)
 
   when (rFirstNonspace rule) $ do
-    firstNonspace <- gets firstNonspaceColumn
-    col <- gets column
+    !firstNonspace <- gets firstNonspaceColumn
+    !col <- gets column
     guard (firstNonspace == Just col)
 
   oldstate <- if rLookahead rule
                  then Just <$> get -- needed for lookahead rules
                  else return Nothing
 
+  -- reset regex captures
+  modify $ \st -> st{ captures = Captures mempty }
+
   let attr = rAttribute rule
   mbtok <- case rMatcher rule of
                 DetectChar c -> withAttr attr $ detectChar (rDynamic rule) c inp
@@ -303,7 +334,8 @@ tryRule rule inp = do
                                     stringDetect (rDynamic rule) (rCaseSensitive rule)
                                                  s inp
                 WordDetect s -> withAttr attr $
-                                    wordDetect (rCaseSensitive rule) s inp
+                                    wordDetect (rCaseSensitive rule)
+                                      (rWeakDeliminators rule) s inp
                 LineContinue -> withAttr attr $ lineContinue inp
                 DetectSpaces -> withAttr attr $ detectSpaces inp
                 DetectIdentifier -> withAttr attr $ detectIdentifier inp
@@ -337,8 +369,11 @@ tryRule rule inp = do
 
   info $ takeWhile (/=' ') (show (rMatcher rule)) ++ " MATCHED " ++ show mbtok'
   doContextSwitches (rContextSwitch rule)
+  -- Add any captures to the context on top of the stack
+  addCaptures
   return mbtok'
 
+
 withAttr :: TokenType -> TokenizerM Text -> TokenizerM (Maybe Token)
 withAttr tt p = do
   res <- p
@@ -346,9 +381,9 @@ withAttr tt p = do
      then return Nothing
      else return $ Just (tt, res)
 
-wordDetect :: Bool -> Text -> ByteString -> TokenizerM Text
-wordDetect caseSensitive s inp = do
-  wordBoundary inp
+wordDetect :: Bool -> Set.Set Char -> Text -> ByteString -> TokenizerM Text
+wordDetect caseSensitive weakDelims s inp = do
+  wordBoundary weakDelims inp
   t <- decodeBS $ UTF8.take (Text.length s) inp
   -- we assume here that the case fold will not change length,
   -- which is safe for ASCII keywords and the like...
@@ -361,7 +396,7 @@ wordDetect caseSensitive s inp = do
   let d = case UTF8.uncons rest of
                Nothing    -> '\n'
                Just (x,_) -> x
-  guard $ isWordBoundary c d
+  guard $ isWordBoundary weakDelims c d
   takeChars (Text.length t)
 
 stringDetect :: Bool -> Bool -> Text -> ByteString -> TokenizerM Text
@@ -421,6 +456,7 @@ includeRules mbattr (syn, con) inp = do
            Text.unpack con ++ "##" ++ Text.unpack syn
        Just c   -> do
          mbtok <- msum (map (\r -> tryRule r inp) (cRules c))
+         modify $ \st -> st{ captures = Captures mempty }
          return $ case (mbtok, mbattr) of
                     (Just (NormalTok, xs), Just attr) -> Just (attr, xs)
                     _                                 -> mbtok
@@ -514,26 +550,19 @@ regExpr :: Bool -> RE -> ByteString -> TokenizerM Text
 regExpr dynamic re inp = do
   -- return $! traceShowId $! (reStr, inp)
   let reStr = reString re
-  when (BS.take 2 reStr == "\\b") $ wordBoundary inp
-  compiledREs <- gets compiledRegexes
-  regex <- case Map.lookup re compiledREs of
-              Nothing -> do
-                cre <- case compileRegex (reCaseSensitive re) reStr of
-                         Right r  -> return r
-                         Left e   -> throwError $
-                           "Error compiling regex " ++
-                            UTF8.toString reStr ++ ": " ++ e
-                modify $ \st -> st{ compiledRegexes =
-                      Map.insert re cre (compiledRegexes st) }
-                return cre
-              Just cre -> return cre
+  when (BS.take 2 reStr == "\\b") $ wordBoundary mempty inp
+  regex <- case compileRE re of
+            Right r  -> return r
+            Left e   -> throwError $
+              "Error compiling regex " ++
+              UTF8.toString reStr ++ ": " ++ e
   regex' <- if dynamic
                then subDynamic regex
                else return regex
   case matchRegex regex' inp of
         Just (matchedBytes, capts) -> do
           unless (null capts) $
-             modify $ \st -> st{ captures =
+            modify $ \st -> st{ captures = Captures $
                                   IntMap.map (toSlice inp) capts }
           takeChars (UTF8.length matchedBytes)
         _ -> mzero
@@ -541,16 +570,18 @@ regExpr dynamic re inp = do
 toSlice :: ByteString -> (Int, Int) -> ByteString
 toSlice bs (off, len) = BS.take len $ BS.drop off bs
 
-wordBoundary :: ByteString -> TokenizerM ()
-wordBoundary inp = do
+wordBoundary :: Set.Set Char -> ByteString -> TokenizerM ()
+wordBoundary weakDelims inp = do
   case UTF8.uncons inp of
        Nothing -> return ()
        Just (d, _) -> do
          c <- gets prevChar
-         guard $ isWordBoundary c d
+         guard $ isWordBoundary weakDelims c d
 
-isWordBoundary :: Char -> Char -> Bool
-isWordBoundary c d = isWordChar c /= isWordChar d
+isWordBoundary :: Set.Set Char -> Char -> Char -> Bool
+isWordBoundary weakDelims c d =
+  (isWordChar c || c `Set.member` weakDelims) /=
+  (isWordChar d || d `Set.member` weakDelims)
 
 decodeBS :: ByteString -> TokenizerM Text
 decodeBS bs = case decodeUtf8' bs of
@@ -578,13 +609,6 @@ subDynamic (AssertNegative dir r) =
   AssertNegative dir <$> subDynamic r
 subDynamic x = return x
 
-getCapture :: Int -> TokenizerM Text
-getCapture capnum = do
-  capts <- gets captures
-  case IntMap.lookup capnum capts of
-     Nothing -> throwError $ "Capture " <> show capnum <> " not defined!"
-     Just x  -> decodeBS x
-
 keyword :: KeywordAttr -> WordSet Text -> ByteString -> TokenizerM Text
 keyword kwattr kws inp = do
   prev <- gets prevChar
@@ -599,11 +623,13 @@ keyword kwattr kws inp = do
 
 normalizeHighlighting :: [Token] -> [Token]
 normalizeHighlighting [] = []
-normalizeHighlighting ((t,x):xs)
+normalizeHighlighting ((!t,!x):xs)
   | Text.null x = normalizeHighlighting xs
   | otherwise =
-    (t, Text.concat (x : map snd matches)) : normalizeHighlighting rest
+    (t, matchedText) : normalizeHighlighting rest
     where (matches, rest) = span (\(z,_) -> z == t) xs
+          !matchedText = Text.concat (x : map snd matches)
+
 
 parseCStringChar :: ByteString -> TokenizerM Text
 parseCStringChar inp = do
@@ -635,7 +661,7 @@ pCChar = do
 
 parseInt :: ByteString -> TokenizerM Text
 parseInt inp = do
-  wordBoundary inp
+  wordBoundary mempty inp
   case A.parseOnly (A.match (pHex <|> pOct <|> pDec)) inp of
        Left _      -> mzero
        Right (r,_) -> takeChars (BS.length r) -- assumes ascii
@@ -643,12 +669,11 @@ parseInt inp = do
 pDec :: A.Parser ()
 pDec = do
   mbMinus
-  _ <- A.takeWhile1 (A.inClass "0-9")
-  guardWordBoundary
+  void $ A.takeWhile1 (A.inClass "0-9")
 
 parseOct :: ByteString -> TokenizerM Text
 parseOct inp = do
-  wordBoundary inp
+  wordBoundary mempty inp
   case A.parseOnly (A.match pHex) inp of
        Left _      -> mzero
        Right (r,_) -> takeChars (BS.length r) -- assumes ascii
@@ -659,11 +684,11 @@ pOct = do
   _ <- A.char '0'
   _ <- A.satisfy (A.inClass "Oo")
   _ <- A.takeWhile1 (A.inClass "0-7")
-  guardWordBoundary
+  return ()
 
 parseHex :: ByteString -> TokenizerM Text
 parseHex inp = do
-  wordBoundary inp
+  wordBoundary mempty inp
   case A.parseOnly (A.match pHex) inp of
        Left _      -> mzero
        Right (r,_) -> takeChars (BS.length r) -- assumes ascii
@@ -674,14 +699,7 @@ pHex = do
   _ <- A.char '0'
   _ <- A.satisfy (A.inClass "Xx")
   _ <- A.takeWhile1 (A.inClass "0-9a-fA-F")
-  guardWordBoundary
-
-guardWordBoundary :: A.Parser ()
-guardWordBoundary = do
-  mbw <- A.peekChar
-  case mbw of
-       Just c  ->  guard $ isWordBoundary '0' c
-       Nothing -> return ()
+  return ()
 
 mbMinus :: A.Parser ()
 mbMinus = (() <$ A.char '-') <|> return ()
@@ -691,7 +709,7 @@ mbPlusMinus = () <$ A.satisfy (A.inClass "+-") <|> return ()
 
 parseFloat :: ByteString -> TokenizerM Text
 parseFloat inp = do
-  wordBoundary inp
+  wordBoundary mempty inp
   case A.parseOnly (A.match pFloat) inp of
        Left _      -> mzero
        Right (r,_) -> takeChars (BS.length r)  -- assumes all ascii
